import inspect
from typing import Any, Optional, Type, Dict, Tuple


class Scope:
    def __init__(
        self,
        name: str,
        parent: "Optional[Scope]" = None,
    ):
        self.name = name
        self.parent = parent

        self.file = None
        self.qualified_name = None
        if parent is not None:
            self.file = parent.file
            self.qualified_name = f"{parent.qualified_name}.{name}"

        self.variables: Dict[str, Type] = {}
        self.types = {}
        self.callables = {}

    def add_variable(self, name: str, annotation: Type) -> None:
        self.variables[name] = annotation

    def resolve_variable(self, name: str) -> Optional[Type]:
        annotation = self.variables.get(name, None)
        if annotation is None and self.parent is not None:
            annotation = self.parent.resolve_variable(name)

        return annotation

    def add_callable(self, name: str, obj: Any, cb: inspect.Signature) -> None:
        self.callables[name] = (obj, cb)

    def resolve_callable(self, name: str) -> Optional[Tuple[Any, inspect.Signature]]:
        obj, cb = self.callables.get(name, (None, None))

        if cb is None and self.parent is not None:
            return self.parent.resolve_callable(name)

        return obj, cb

    @staticmethod
    def _format_dict(dictionary: Dict[str, Any]) -> str:
        if dictionary:
            items = ",\n    ".join(dictionary.keys())
            return f"\n    {items}\n  "

        return ""

    def __str__(self) -> str:
        return (
            f"<Scope '{self.qualified_name}'\n"
            f"  variable=[{self._format_dict(self.variables)}]\n"
            f"  types=[]\n"
            f"  callables=[{self._format_dict(self.callables)}]\n"
            f">"
        )


# TODO: tests
def parse_scope(
    container: Any,
    parent_scope: Optional[Scope] = None,
) -> Scope:
    """Populate scope with all existing symbols."""
    scope = Scope(container.__name__, parent_scope)

    for symbol, obj in container.__dict__.items():
        if callable(obj):
            signature = inspect.signature(obj)

            if hasattr(obj, "__name__"):
                name = obj.__name__
            # This if for objects from the typing module,
            # such as Optional or Union
            elif hasattr(obj, "_name"):
                name = obj._name
            else:
                raise NotImplementedError(
                    f"name retrieval for callable '{obj}' is not implemented"
                )

            scope.add_callable(name, obj, signature)

    return scope



import logging
import os
import sys
from typing import List

from typpy.discover_files import find_files
from typpy.output import print_errors
from typpy.type_checker import TypeChecker


# TODO: test, docstring
def run(args: List[str] = None) -> None:
    args = args or sys.argv

    logging.basicConfig(level=os.environ.get("LOG_LEVEL", "INFO").upper())

    # TODO: move to proper function
    files = find_files([args[1]])

    type_checker = TypeChecker()
    errors = type_checker.check_files(files)

    if errors:
        print_errors(errors)
        sys.exit(1)


if __name__ == "__main__":
    run(sys.argv)



import ast
import inspect
import logging
import sys
import types
from contextlib import contextmanager
from dataclasses import dataclass
from importlib import import_module
from pathlib import Path
from typing import Iterable, Generator, Any, List

from typpy.error import TypingError
from typpy.expression import check_expression
from typpy.resolver import Resolver
from typpy.scope import Scope, parse_scope
from typpy.statement import check_assignment


@dataclass(frozen=True)
class FileModule:
    qualified_name: str
    path: Path
    module: types.ModuleType


# TODO: docstring, tests
class TypeChecker:
    def __init__(self):
        self.resolver = Resolver()

    def check_files(self, files: Iterable[Path]) -> List[TypingError]:
        pending_files = set(files)
        processed_files = set()

        errors = []
        while pending_files:
            file = pending_files.pop()
            new_errors = self._check_file(file)
            errors.extend(new_errors)
            processed_files.add(file)

        return errors

    @contextmanager
    def import_module(self, path: Path) -> Generator[FileModule, None, None]:
        module_name, containing_path = self.resolver.resolve(path)

        sys.path.append(str(containing_path))
        try:
            module = import_module(module_name)

            yield FileModule(
                qualified_name=module_name,
                path=path,
                module=module,
            )
        finally:
            sys.path.remove(str(containing_path))

    def _check_file(self, path: Path) -> List[TypingError]:
        with self.import_module(path) as module:
            scope = parse_scope(module.module)
            # TODO: better way to bootstrap?
            # Bootstrap scope qualified name
            scope.file = path
            scope.qualified_name = module.qualified_name
            return self._check_scope_typing(module.module, scope)

    def _check_scope(
        self,
        obj: Any,
        parent_scope: Scope,
    ) -> List[TypingError]:
        # Parse scope before everything else so that all symbols
        # are already parsed and can be referenced without assuming
        # they come before the checked statement.
        scope = parse_scope(obj, parent_scope)
        return self._check_scope_typing(obj, scope)

    def _check_scope_typing(self, obj: Any, scope: Scope) -> List[TypingError]:
        tree = ast.parse(inspect.getsource(obj))

        if not isinstance(obj, types.ModuleType):
            tree = tree.body[0]

        errors = []
        for stmt in tree.body:
            new_errors = self._check_statement(stmt, scope)
            errors.extend(new_errors)

        # Check sub-scopes
        # TODO: clean-up
        for cb in scope.callables.values():
            new_errors = self._check_scope(cb[0], scope)
            errors.extend(new_errors)

        logging.debug("obj: %s - scope: %s" % (obj.__name__, scope))

        return errors

    def _check_statement(self, stmt: ast.stmt, scope: Scope) -> List[TypingError]:
        logging.debug("stmt: %s" % stmt)

        # These two are already evaluated by the module import
        # and are retrieved dynamically instead of parsed.
        if isinstance(stmt, (ast.ClassDef, ast.FunctionDef)):
            return []

        errors = []
        if isinstance(stmt, ast.If):
            errors = check_expression(stmt.test, scope)
        elif isinstance(stmt, (ast.Assign, ast.AnnAssign)):
            errors = check_assignment(stmt, scope)
        elif isinstance(stmt, ast.Expr):
            errors = check_expression(stmt.value, scope)

        for sub_stmt in getattr(stmt, "body", []):
            new_errors = self._check_statement(sub_stmt, scope)
            errors.extend(new_errors)

        return errors



from typing import Type, Union, Tuple, Any, Optional
from inspect import isclass


# TODO: test
def is_subtype(act_type: Optional[Type], exp_type: Optional[Type]) -> bool:
    act_type = type(None) if act_type is None else act_type
    exp_type = type(None) if exp_type is None else exp_type

    # optimization for common use case
    if act_type is exp_type:
        return True

    # SPECIAL CASES
    # =============
    # if the expected type is Any, the result is always True
    if exp_type is Any:
        return True
    # For numeric types, we follow PEP 484, and not
    # the class hierarchy (e.g. issubclass(int, float)
    # is False but we return True)
    elif exp_type is float and act_type in (int, bool):
        return True
    elif exp_type is complex and act_type in (float, int, bool):
        return True

    exp_origin_type = getattr(exp_type, "__origin__", None)
    if exp_origin_type is not None:
        # This if for objects from the typing module,
        # such as Optional or Union
        if exp_origin_type is Union:
            return _check_union(act_type, exp_type)
        elif exp_origin_type in (Tuple, tuple):
            return _check_tuple(act_type, exp_type)
    elif isclass(act_type):
        return issubclass(act_type, exp_type)
    else:
        # This happens when act_type is a Tuple or Union
        # and ext_type is not. This is always False.
        return False

    raise NotImplementedError(f"is_subtype() for type {exp_type} is not implemented")


def _check_union(act_type: Type, exp_type: Union) -> bool:
    # If the act_type is an Union, check that act_type is a subset
    if getattr(act_type, "__origin__", None) is Union:
        union_types = set(exp_type.__args__)
        return all(t in union_types for t in act_type.__args__)

    # Otherwise, check it act_type is contained in the Union
    for union_type in exp_type.__args__:
        if is_subtype(act_type, union_type):
            return True

    return False


def _check_tuple(act_type: Type, exp_type: Tuple) -> bool:
    if getattr(act_type, "__origin__", None) in (tuple, Tuple):
        return all(
            is_subtype(act, exp)
            for act, exp in zip(act_type.__args__, exp_type.__args__)
        )

    return False



import ast
from dataclasses import dataclass
from pathlib import Path

from typpy.scope import Scope


@dataclass(frozen=True)
class TypingError:
    file: Path
    line_number: int
    column_number: int
    message: str

    @classmethod
    def from_scope_stmt(
        cls,
        scope: Scope,
        stmt: ast.stmt,
        message: str,
    ) -> "TypingError":
        return cls(
            file=scope.file,
            line_number=stmt.lineno,
            column_number=stmt.col_offset,
            message=message,
        )



import sys
from typing import List

from typpy.error import TypingError


# TODO: tests
def print_errors(errors: List[TypingError]) -> None:
    last_file = None

    num_files = 0
    for error in errors:
        if error.file != last_file:
            print(error.file, file=sys.stderr)
            last_file = error.file
            num_files += 1

        print(f"  {error.file}:{error.line_number}: {error.message}", file=sys.stderr)

    print()
    print(f"Found {len(errors)} errors in {num_files} file")



from functools import lru_cache
from pathlib import Path
from typing import Tuple


class Resolver:
    """Resolve paths into fully qualified module names,
    caching filesystem access for speed.
    """

    # Cache already resolved paths, to avoid hitting the
    # filesystem many times to resolve the same folders.
    @lru_cache(maxsize=None)
    def is_package(self, path: Path) -> bool:
        return (path / "__init__.py").exists()

    # See comment above
    @lru_cache(maxsize=None)
    def resolve(self, path: Path, _check_path: bool = True) -> Tuple[str, Path]:
        """Resolve paths into fully qualified module names.

        :param path: A path to a file or directory.
        :param _check_path: For internal use.
        :return: A tuple containing a fully qualified
            module name and the path to the folder containing
            the outermost package.
        """
        if _check_path:
            if path.suffix:
                # A file bust be a python source file
                assert path.suffix == ".py"

            # Check that the path is importable. If not, it should
            # not have been passed here to be resolved and indicates
            # a bug elsewhere.
            if path.suffix != ".py" and path.is_dir():
                # A directory should contain a __init__.py file
                assert self.is_package(path)

        if not self.is_package(path.parent):
            # We reached the folder containing the outermost package
            return path.stem, path.parent

        parent_qualified_name, containing_path = self.resolve(
            path.parent, _check_path=False
        )
        return f"{parent_qualified_name}.{path.stem}", containing_path



# Necessary to build wheel
import setuptools
from distutils.core import setup
from pathlib import Path

ROOT = Path(__file__).parent

setup(
    name="typpy",
    version="0.2.0",
    description="Python Type Checker",
    long_description=(ROOT / "README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Francesco Pasa",
    author_email="francescopasa@gmail.com",
    url="https://github.com/frapa/typpy",
    license="GPL3",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Typing :: Typed",
    ],
    packages=["typpy"],
    package_data={
        "typpy": ["py.typed"],
    },
    include_package_data=True,
    entry_points={
        "console_scripts": ["typpy=typpy.__main__:run"],
    },
)



from glob import glob
from pathlib import Path
from typing import List, Iterable


# TODO: test, docstring
def find_files(glob_patterns: List[str]) -> Iterable[Path]:
    for glob_pattern in glob_patterns:
        yield from _expand_glob(glob_pattern)


def _expand_glob(glob_pattern: str) -> Iterable[Path]:
    yield from (Path(path) for path in glob(glob_pattern))
